#include <stdio.h>
#include <asm/io.h>
#include <sys/ioctl.h>
#include <stdlib.h>

#define PARPORT_DATA   0x378
#define PARPORT_STATUS 0x379
#define PARPORT_CTRL   0x37a

#define X_MOTOR 2
#define Y_MOTOR 0
#define Z_MOTOR 1

#define STEP_DELAY 10000
//#define STEP_DELAY 0

typedef unsigned char u8;

static void step(int port, int fwd);
static void jog_screen(void);

/*
 *  STROBE'         CONTROL, 0
 *  D0
 *  D1
 *  D2
 *  D3
 *  D4
 *  D5
 *  D6
 *  D7
 *  AUTOFEED'       CONTROL, 1
 *  INITIALIZE      CONTROL, 2
 *  SELECT INPUT'   CONTROL, 3
 */

u8 states[12];

int LastDirX;

static void write_bits(void)
{
    u8 data = 0;
    u8 ctrl = 0;
    if( states[ 0]) ctrl |= (1<<0);
    if(!states[ 1]) data |= (1<<0); 
    if(!states[ 2]) data |= (1<<1); 
    if(!states[ 3]) data |= (1<<2); 
    if(!states[ 4]) data |= (1<<3); 
    if(!states[ 5]) data |= (1<<4); 
    if(!states[ 6]) data |= (1<<5); 
    if(!states[ 7]) data |= (1<<6); 
    if(!states[ 8]) data |= (1<<7); 
    if( states[ 9]) ctrl |= (1<<1);
    if(!states[10]) ctrl |= (1<<2);
    if( states[11]) ctrl |= (1<<3);

    outb(ctrl, PARPORT_CTRL);
    outb(data, PARPORT_DATA);
//    printf("write: ctrl=%02x data=%02x\r\n", ctrl, data);
}

void mill_up(void)
{
    // up by 0.050"
    int i;
    usleep(10000);
    for(i = 0; i < 50; i++) {
        memset(states, 0, 12);
        step(Z_MOTOR, 0);
        usleep(STEP_DELAY);
        write_bits();
    }
    memset(states, 0, 12);
    write_bits();
}

void mill_down(void)
{
    // down by 0.050"
    int i;
    usleep(10000);
    for(i = 0; i < 50; i++) {
        memset(states, 0, 12);
        step(Z_MOTOR, 1);
        usleep(STEP_DELAY);
        write_bits();
    }
    memset(states, 0, 12);
    write_bits();
}

void mill_nextsize(void)
{
}

void mill_drill(void)
{
    int i;
    // down by 0.100"
    usleep(10000);
    for(i = 0; i < 100; i++) {
        memset(states, 0, 12);
        step(Z_MOTOR, 1);
        usleep(STEP_DELAY);
        write_bits();
    }
    memset(states, 0, 12);
    write_bits();
    // now up by 0.100"
    usleep(10000);
    for(i = 0; i < 100; i++) {
        memset(states, 0, 12);
        step(Z_MOTOR, 0);
        usleep(STEP_DELAY);
        write_bits();
    }
    memset(states, 0, 12);
    write_bits();
    usleep(10000);
}

void mill_now_drilling(void)
{
    int i;
    printf("preparing for tool change\n");
    // now up by 1.000"
    usleep(10000);
    for(i = 0; i < 1000; i++) {
        memset(states, 0, 12);
        step(Z_MOTOR, 0);
        usleep(STEP_DELAY);
        write_bits();
    }
    printf("insert drill and position over pad0 (white in preview)\n");
    jog_screen();
    printf("now drilling\n");
}

#define abs(x) ((x) > 0 ? (x) : -(x))
void mill_translate(int dx, int dy)
{
    int i;
    int maj, min;
    int magX = abs(dx);
    int magY = abs(dy);
    int dirX = (dx > 0) ? 0 : 1;
    int dirY = (dy > 0) ? 0 : 1;

    usleep(10000);

    if(dx != 0) {
        if(LastDirX != dirX) {
            for(i = 0; i < 16; i++) {
                step(X_MOTOR, dirX);
                usleep(STEP_DELAY);
                write_bits();
            }
            usleep(STEP_DELAY);
        }
        LastDirX = dirX;
    }

    if(magX > magY) {
        for(maj = 0, min = 0; maj < magX || min < magY; maj++) {
            if(min < (maj * magY + (magX / 2)) / magX) {
                step(Y_MOTOR, dirY);
                min++;
            }
            if(maj < magX) {
                step(X_MOTOR, dirX);
            }
            usleep(STEP_DELAY);
            write_bits();
        }
    } else {
        for(maj = 0, min = 0; maj < magY || min < magX; maj++) {
            if(min < (maj * magX + (magY / 2)) / magY) {
                step(X_MOTOR, dirX);
                min++;
            }
            if(maj < magY) {
                step(Y_MOTOR, dirY);
            }
            usleep(STEP_DELAY);
            write_bits();
        }
    }
    usleep(STEP_DELAY);
    memset(states, 0, 12);
    write_bits();
}

static void step(int port, int fwd)
{
    static int indices[3];

    if(fwd) {
        indices[port]++;
        if(indices[port] > 3) indices[port] = 0;
    } else {
        indices[port]--;
        if(indices[port] < 0) indices[port] = 3;
    }
    memset(states + (4*port), 0, 4);
    switch(indices[port]) {
        case 0:
            states[(4*port) + 0] = 1;
            states[(4*port) + 3] = 1;
            break;
        case 1:
            states[(4*port) + 0] = 1;
            states[(4*port) + 1] = 1;
            break;
        case 2:
            states[(4*port) + 1] = 1;
            states[(4*port) + 2] = 1;
            break;
        case 3:
            states[(4*port) + 2] = 1;
            states[(4*port) + 3] = 1;
            break;
    }
}

static void jog_screen(void)
{
    int arg;
    int xdir = -1;
    int ydir = -1;
    int zdir = -1;

    int stepZ0 = 0;
    int stepZ1 = 0;
    int stepY0 = 0;
    int stepX1 = 0;
    int stepX0 = 0;
    int stepY1 = 0;

    printf("jog mode; q to exit\n");
    system("stty raw -echo");
	for(;;) {
        ioctl(0, FIONREAD, &arg);
        if(arg != 0) {
            int c = getchar();
            switch(c) {
                case 'q':
                    system("stty sane");
                    memset(states, 0, 12);
                    write_bits();
                    return;

                case 'W': stepX1 = 1; LastDirX = 1; break;
                case 'X': stepX0 = 1; LastDirX = 0; break;

                case 'w': xdir =  1; LastDirX = 1; break;
                case 's': xdir = -1; break;
                case 'x': xdir =  0; LastDirX = 0; break;

                case 'E': stepY1 = 1; break;
                case 'C': stepY0 = 1; break;

                case 'e': ydir =  1; break;
                case 'd': ydir = -1; break;
                case 'c': ydir =  0; break;

                case 'R': stepZ1 = 1; break;
                case 'V': stepZ0 = 1; break;

                case 'r': zdir =  1; break;
                case 'f': zdir = -1; break;
                case 'v': zdir =  0; break;

                default:
                    printf("unknown '%c'\r\n", c);
                    break;
            }
        }
        memset(states, 0, 12);
        if(xdir >= 0) step(X_MOTOR, xdir);
        if(ydir >= 0) step(Y_MOTOR, ydir);
        if(zdir >= 0) step(Z_MOTOR, zdir);
        if(stepZ1) step(Z_MOTOR, 1);
        if(stepZ0) step(Z_MOTOR, 0);
        if(stepY1) step(Y_MOTOR, 1);
        if(stepY0) step(Y_MOTOR, 0);
        if(stepX1) step(X_MOTOR, 1);
        if(stepX0) step(X_MOTOR, 0);
        stepZ1 = 0;
        stepZ0 = 0;
        stepY1 = 0;
        stepY0 = 0;
        stepX1 = 0;
        stepX0 = 0;
        write_bits();
        usleep(10000);
    }
}

int main(void)
{
    int xp, yp;

    ioperm(PARPORT_DATA, 3, 1);

    if(0) {
        for(yp = 0; yp < 3000;) {
            mill_translate(5000, 0);
            mill_translate(0, 180);
            yp += 300;
            printf("yp = %d\n", yp);
            getchar();
            mill_translate(-5000, 0);
            mill_translate(0, 180);
            yp += 300;
            printf("yp = %d\n", yp);
            getchar();
        }
        return;
    }

    if(0) {
        int i;
        mill_up();
        for(i = 0; i < 15; i++) {
            memset(states, 0, 12);
            step(Z_MOTOR, 1);
            usleep(STEP_DELAY);
            write_bits();
        }
        mill_translate(25, 25);
        mill_from_file("test.scr");
        return;
    }

    jog_screen();
}
