#include <p16f877.inc>

radix dec

  __CONFIG _CP_OFF & _XT_OSC & _PWRTE_ON  & _WDT_OFF & _BODEN_OFF & _LVP_OFF


RESET_V		EQU	0x0000
ISR_V		EQU	0x0004
OSC_FREQ	EQU	D'4000000'

LedOn		macro
	bsf	PORTD,	3
	endm

LedOff		macro
	bcf	PORTD,	3
	endm

LedBlink	macro
	btfss	PORTD,	3
	goto	$ + 3
	bcf	PORTD,	3
	goto	$ + 2
	bsf	PORTD,	3
	endm

PutChar		macro	c
	movlw	c
	call	putChar
	endm

NewLine		macro
	PutChar	'\r'
	PutChar	'\n'
	endm

Preload		macro	stepFn
	local i = 0
	while i < 6
		call	stepFn
		call	waitBetweenSteps
	i += 1
	endw
	endm

cblock 0x20
	wTemp		; temps for ISR context; wTemp must appear in all banks
	pclathTemp
	statusTemp
	ivar, jvar	; counters for waitBetweenChars, waitBetweenSteps
	svar, tvar	; for putHex
	rxChar		; char received on serial
	steps		; step count when moving z axis
	millSteps	; number of steps to take when doing mill up/down
	drillSteps	; number of steps to take when drilling
	defaultMinPer	; default minimum period between steps
	period		; period between steps; we will ramp this!
	xPos
	yPos
	zPos
	writeTemp
	jogging		; map of which way we're jogging x, y, and z
	xTotal:2	; total number of x, y steps that we want to take
	yTotal:2
	xyDir
	maxTotal:2	; total number of steps that the larger direction takes
	minTotal:2	; total number of steps that the smaller direction takes
	maxSoFar:2
	minSoFar:2
	maxAcc:4	; minTotal*maxSoFar
	minAcc:4	; maxTotal*minSoFar
	maxStepFn, maxStepFnHigh
	minStepFn, minStepFnHigh
endc

; which bit does what in jogging
JOG_X_FWD	equ	0
JOG_X_BACK	equ	1
JOG_Y_FWD	equ	2
JOG_Y_BACK	equ	3
JOG_Z_FWD	equ	4
JOG_Z_BACK	equ	5
JOG_NONE	equ	6

; which bit does what in xyDir
X_DIR_FWD	equ	0
Y_DIR_FWD	equ	1

; we end up here after reset
	org	0
	goto	setUpHardware
	org	4
; we end up here when we interrupt
	goto	isr

isr
	; context save; from datasheet
	movwf	wTemp
	swapf	STATUS,		W
	clrf	STATUS
	movwf	statusTemp
	movf	PCLATH,		W
	movwf	pclathTemp
	clrf	PCLATH

	; let's see who interrupted
	btfsc	INTCON,		T0IF
	goto	timer0Expired

	; strange; why did we get called?
	goto	endIsr

timer0Expired
	banksel	INTCON
	bcf	INTCON,		T0IF

	movlw	d'80'
	addwf	TMR0,		F

	; now let's see what to jog
	btfsc	jogging,	JOG_X_FWD
	call	xStepFwd
	btfsc	jogging,	JOG_X_BACK
	call	xStepBack
	btfsc	jogging,	JOG_Y_FWD
	call	yStepFwd
	btfsc	jogging,	JOG_Y_BACK
	call	yStepBack
	btfsc	jogging,	JOG_Z_FWD
	call	zStepFwd
	btfsc	jogging,	JOG_Z_BACK
	call	zStepBack

	goto	endIsr

endIsr
	; context restore; from datasheet
	movf	pclathTemp,	W
	movwf	PCLATH
	swapf	statusTemp,	W
	movwf	STATUS
	swapf	wTemp,		F
	swapf	wTemp,		W
	retfie

setUpHardware

	nop
	nop
	bcf	STATUS,		RP0	
	bcf	STATUS,		RP1
	bsf	STATUS,		RP0
	; For the stepper control lines, and for the LED
	clrf	TRISB ^ 0x80
	clrf	TRISD ^ 0x80

	; UART pins must be tri-stated, others are unused (so tri-state 'em)
	movlw	0xff
	movwf	TRISC ^ 0x80

	; Initialize SPBRG register with appropriate value for 2400 bps
	movlw	d'25'
	movwf	SPBRG ^ 0x80
	; Enable UART by clearing bit SYNC and setting bit SPEN
	bcf	TXSTA ^ 0x80,	SYNC
	bcf	STATUS,		RP0
	bsf	RCSTA,		SPEN
	; Enable transmission by setting bit TXEN
	bsf	STATUS,		RP0
	bsf	TXSTA ^ 0x80,	TXEN

	bcf	STATUS,		RP0

	; Now let's get timer0 firing for jog mode
	bsf	INTCON,		T0IE

	banksel OPTION_REG
	bcf	OPTION_REG ^ 0x80,	T0CS	; 0 = CLKOUT
	bcf	OPTION_REG ^ 0x80,	PSA	; 0 = prescaler for timer0
	bsf	OPTION_REG ^ 0x80,	0	; 101 means 1:64
	bcf	OPTION_REG ^ 0x80,	1
	bsf	OPTION_REG ^ 0x80,	2

	; No sense powering the steppers unless we need to
	banksel	PORTD
	clrf	PORTD
	clrf	PORTB

	; reasonable defaults for lots of things
	movlw	d'50'
	movwf	millSteps
	movlw	d'250'
	movwf	drillSteps
	movlw	0x40
	movwf	defaultMinPer

	call	placeRampAtBeginning

	; And let's make sure we know where our steppers are
	LedOn
	clrf	xPos
	Preload	xStepFwd
	Preload	xStepBack
	call	xRelease

	LedOff
	clrf	yPos
	Preload	yStepFwd
	Preload	yStepBack
	call	yRelease

	LedOn
	clrf	zPos
	Preload	zStepFwd
	Preload	zStepBack
	call	zRelease
	
	; We're alive; tell the computer
	PutChar	'v'
	PutChar	'0'
	PutChar	'.'
	PutChar	'0'

	; interrupts on and let's go!
	bsf	INTCON,		GIE

serialRead
	NewLine
	
	LedOff
	call	getChar
	movwf	rxChar
	LedOn

;;; let's see what they want us to do

	; PING REQUEST
	movf	rxChar,		w
	sublw	'p'
	btfss	STATUS,		Z
	goto	notPing

	PutChar	'p'
	PutChar	'o'
	PutChar	'n'
	PutChar	'g'
	
	goto	serialRead
notPing

	; one of the single-step commands
Step		macro	char, jumpIfNot, callIf, release
		
	movf	rxChar,		W
	sublw	char
	btfss	STATUS,		Z
	goto	jumpIfNot

	call	callIf
	call	waitBetweenSteps
	call	release

	goto	serialRead
jumpIfNot
	endm

	Step 'W', notStepXFwd,	xStepFwd,	xRelease
	Step 'X', notStepXBack, xStepBack,	xRelease
	Step 'E', notStepYFwd,	yStepFwd,	yRelease
	Step 'C', notStepYBack, yStepBack,	yRelease
	Step 'R', notStepZFwd,	zStepFwd,	zRelease
	Step 'V', notStepZBack, zStepBack,	zRelease

	; one of the jog commands
Jog		macro	char, jumpIfNot, setIf, clear0, clear1, callCmd
		
	movf	rxChar,		W
	sublw	char
	btfss	STATUS,		Z
	goto	jumpIfNot

	bcf	jogging,	clear0
	bcf	jogging,	clear1
	bsf	jogging,	setIf

	callCmd

	goto	serialRead
jumpIfNot
	endm

	Jog 'w', notJogXFwd,	JOG_X_FWD,	JOG_X_FWD,	JOG_X_BACK, nop
	Jog 'x', notJogXBack,	JOG_X_BACK,	JOG_X_FWD,	JOG_X_BACK, nop
	Jog 'e', notJogYFwd,	JOG_Y_FWD,	JOG_Y_FWD,	JOG_Y_BACK, nop
	Jog 'c', notJogYBack,	JOG_Y_BACK,	JOG_Y_FWD,	JOG_Y_BACK, nop
	Jog 'r', notJogZFwd,	JOG_Z_FWD,	JOG_Z_FWD,	JOG_Z_BACK, nop
	Jog 'v', notJogZBack,	JOG_Z_BACK,	JOG_Z_FWD,	JOG_Z_BACK, nop
	Jog 's', notJogXNone, JOG_NONE, JOG_X_FWD, JOG_X_BACK, call xRelease
	Jog 'd', notJogYNone, JOG_NONE, JOG_Y_FWD, JOG_Y_BACK, call yRelease
	Jog 'f', notJogZNone, JOG_NONE,	JOG_Z_FWD, JOG_Z_BACK, call zRelease

	; MILL DOWN
	movf	rxChar,		w
	sublw	'l'
	btfss	STATUS,		Z
	goto	notMillDown

	call	placeRampAtBeginning

	movf 	millSteps,	W
	movwf	steps
millDownStep
	call	zStepFwd
	call	waitBetweenSteps
	decfsz	steps,		F
	goto	millDownStep
	
	call	waitBetweenSteps
	call	waitBetweenSteps
	call	waitBetweenSteps
	call	zRelease

	goto	serialRead
notMillDown

	; MILL UP
	movf	rxChar,		w
	sublw	'o'
	btfss	STATUS,		Z
	goto	notMillUp

	call	placeRampAtBeginning

	movf	millSteps,	W
	movwf	steps
millUpStep
	call	zStepBack
	call	waitBetweenSteps
	decfsz	steps,		F
	goto	millUpStep

	call	waitBetweenSteps
	call	waitBetweenSteps
	call	waitBetweenSteps
	call	zRelease

	goto	serialRead
notMillUp

	; DRILL
	movf	rxChar,		w
	sublw	'i'
	btfss	STATUS,		Z
	goto	notDrill

	call	placeRampAtBeginning

	movf	drillSteps,	W
	movwf	steps
drillDownStep
	call	zStepFwd
	call	waitBetweenSteps
	decfsz	steps,		F
	goto	drillDownStep

	call	waitBetweenSteps
	call	waitBetweenSteps
	call	waitBetweenSteps

	call	placeRampAtBeginning

	movf	drillSteps,	W
	movwf	steps
drillUpStep
	call	zStepBack
	call	waitBetweenSteps
	decfsz	steps,		F
	goto	drillUpStep

	call	waitBetweenSteps
	call	waitBetweenSteps
	call	waitBetweenSteps
	call	zRelease

	goto	serialRead
notDrill

	; SET DRILL STEPS
	movf	rxChar,		w
	sublw	'I'
	btfss	STATUS,		Z
	goto	notDrillSteps

	call	getChar
	movwf	drillSteps

	goto	serialRead
notDrillSteps

	; SET MILL STEPS
	movf	rxChar,		w
	sublw	'O'
	btfss	STATUS,		Z
	goto	notMillSteps

	call	getChar
	movwf	millSteps

	goto	serialRead
notMillSteps

	; SET DEFAULT PERIOD
	movf	rxChar,		w
	sublw	'a'
	btfss	STATUS,		Z
	goto	notSetDefaultPeriod

	call	getChar
	movwf	defaultMinPer

	goto	serialRead
notSetDefaultPeriod

	; TRANSLATE (the big one!)
	movf	rxChar,		w
	sublw	't'
	btfss	STATUS,		Z
	goto	notTranslate

	call	getChar
	movwf	xyDir

	call	getChar
	movwf	xTotal+1
	call	getChar
	movwf	xTotal+0

	call	getChar
	movwf	yTotal+1
	call	getChar
	movwf	yTotal+0

	clrf	maxSoFar+0
	clrf	maxSoFar+1
	clrf	minSoFar+0
	clrf	minSoFar+1

	clrf	maxAcc+0
	clrf	maxAcc+1
	clrf	maxAcc+2
	clrf	maxAcc+3

	clrf	minAcc+0
	clrf	minAcc+1
	clrf	minAcc+2
	clrf	minAcc+3

	call	placeRampAtBeginning

	; now see whether x > y or y > x (would work either way if equal)

	movf	xTotal+1,	W
	subwf	yTotal+1,	W
	btfsc	STATUS,		Z
	goto	lowDecides
	btfsc	STATUS,		C
	goto	yIsHigher
	goto	xIsHigher

lowDecides
	movf	xTotal+0,	W
	subwf	yTotal+0,	W
	btfsc	STATUS,		C
	goto	yIsHigher
	goto	xIsHigher

yIsHigher
	movf	yTotal+0,	W
	movwf	maxTotal+0
	movf	yTotal+1,	W
	movwf	maxTotal+1
	movf	xTotal+0,	W
	movwf	minTotal+0
	movf	xTotal+1,	W
	movwf	minTotal+1

yIsHigherSetMax
	btfsc	xyDir,		Y_DIR_FWD
	goto	yIsHigherSetMaxFwd

yIsHigherSetMaxBack
	movlw	low yStepBack
	movwf	maxStepFn
	movlw	high yStepBack
	movwf	maxStepFnHigh
	goto	yIsHigherSetMin

yIsHigherSetMaxFwd
	movlw	low yStepFwd
	movwf	maxStepFn
	movlw	high yStepFwd
	movwf	maxStepFnHigh
	goto	yIsHigherSetMin

yIsHigherSetMin
	btfsc	xyDir,		X_DIR_FWD
	goto	yIsHigherSetMinFwd

yIsHigherSetMinBack
	movlw	low xStepBack
	movwf	minStepFn
	movlw	high xStepBack
	movwf	minStepFnHigh
	goto	doTranslate

yIsHigherSetMinFwd
	movlw	low xStepFwd
	movwf	minStepFn
	movlw	high xStepFwd
	movwf	minStepFnHigh
	goto	doTranslate

xIsHigher
	movf	xTotal+0,	W
	movwf	maxTotal+0
	movf	xTotal+1,	W
	movwf	maxTotal+1
	movf	yTotal+0,	W
	movwf	minTotal+0
	movf	yTotal+1,	W
	movwf	minTotal+1

xIsHigherSetMax
	btfsc	xyDir,		X_DIR_FWD
	goto	xIsHigherSetMaxFwd

xIsHigherSetMaxBack
	movlw	low xStepBack
	movwf	maxStepFn
	movlw	high xStepBack
	movwf	maxStepFnHigh
	goto	xIsHigherSetMin

xIsHigherSetMaxFwd
	movlw	low xStepFwd
	movwf	maxStepFn
	movlw	high xStepFwd
	movwf	maxStepFnHigh
	goto	xIsHigherSetMin

xIsHigherSetMin
	btfsc	xyDir,		Y_DIR_FWD
	goto	xIsHigherSetMinFwd

xIsHigherSetMinBack
	movlw	low yStepBack
	movwf	minStepFn
	movlw	high yStepBack
	movwf	minStepFnHigh
	goto	doTranslate

xIsHigherSetMinFwd
	movlw	low yStepFwd
	movwf	minStepFn
	movlw	high yStepFwd
	movwf	minStepFnHigh
	goto	doTranslate

Accumulate32	macro	Dest32,	Src16, doneLso, doneMso
	; first we add the least significant octet of Src16
	movf	Src16+0,	W
	addwf	Dest32+0,	F
	btfss	STATUS,		C
	goto	doneLso
	incf	Dest32+1,	F
	btfss	STATUS,		Z
	goto	doneLso
	incf	Dest32+2,	F
	btfss	STATUS,		Z
	goto	doneLso
	incf	Dest32+3,	F
doneLso
	; and then we add the most significant octet of Src16
	movf	Src16+1,	W
	addwf	Dest32+1,	F
	btfss	STATUS,		C
	goto	doneMso
	incf	Dest32+2,	F
	btfss	STATUS,		Z
	goto	doneMso
	incf	Dest32+3,	F
doneMso
	endm

Inc16		macro	v16
	incf	v16+0,		F
	btfsc	STATUS,		Z
	incf	v16+1,		F
	endm

doTranslate
	; So x is our major (ie. we take an x-step every time). We want to
	; take a y step whenever
	;   (xTaken / xTotal) > (yTaken / yTotal)
	;    xTaken * yTotal  >  yTaken * xTotal
	; We therefore keep a count of xTaken * yTotal and yTaken * xTotal as
	; xTaken and yTotal increase. This has to be a 32 bit value since
	; sqrt(2**24) < total x travel.

GotoIfNotEqual16 macro	SrcA16,	SrcB16, Label
	movf	SrcA16+1,	W
	subwf	SrcB16+1,	W
	btfss	STATUS,		Z
	goto	Label
	movf	SrcA16+0,	W
	subwf	SrcB16+0,	W
	btfss	STATUS,		Z
	goto	Label
	endm

translateStepLoop
	GotoIfNotEqual16 maxSoFar, maxTotal, reallyMaxStep
	goto noMaxStep

reallyMaxStep
	; otherwise we always take a step in the longer direction
	call	maxStep
	; and so we increment 
	Accumulate32	maxAcc,	minTotal,	lab1,	lab2
	Inc16		maxSoFar

noMaxStep
	; now see if we'll take a min step
	movf	maxAcc+3,	W
	subwf	minAcc+3,	W
	btfsc	STATUS,		Z
	goto	octet2Decides
	btfsc	STATUS,		C
	goto	noMinStep
	goto	yesMinStep

octet2Decides
	movf	maxAcc+2,	W
	subwf	minAcc+2,	W
	btfsc	STATUS,		Z
	goto	octet1Decides
	btfsc	STATUS,		C
	goto	noMinStep
	goto	yesMinStep

octet1Decides
	movf	maxAcc+1,	W
	subwf	minAcc+1,	W
	btfsc	STATUS,		Z
	goto	octet0Decides
	btfsc	STATUS,		C
	goto	noMinStep
	goto	yesMinStep

octet0Decides
	movf	maxAcc+0,	W
	subwf	minAcc+0,	W
	btfsc	STATUS,		C
	goto	noMinStep
	goto	yesMinStep

yesMinStep
	GotoIfNotEqual16 minSoFar, minTotal, reallyMinStep
	goto	noMinStep

reallyMinStep
	call	minStep
	Accumulate32	minAcc,	maxTotal,	lab3,	lab4
	Inc16		minSoFar
noMinStep
	call	waitBetweenSteps
	nop

	; now let's see if we're done
	GotoIfNotEqual16 minSoFar, minTotal, translateStepLoop
	GotoIfNotEqual16 maxSoFar, maxTotal, translateStepLoop

	goto	serialRead

;; some helper functions to do our 'function pointer'
minStep
	movf	minStepFnHigh,	W
	movwf	PCLATH
	movf	minStepFn,	W
	movwf	PCL
	; called function returns

maxStep
	movf	maxStepFnHigh,	W
	movwf	PCLATH
	movf	maxStepFn,	W
	movwf	PCL
	; called function returns

notTranslate
	
	
	; unrecognized; indicate error
	PutChar	'e'
	PutChar	'r'
	PutChar	'r'
	PutChar	'o'
	PutChar	'r'

	goto serialRead

;;;	write the appropriate waveforms for the x, y, z motors; we have a
;;;	lookup table so make sure we won't wrap
org	0x400
xWrite
	movf	PORTB,		W
	andlw	0x0f
	movwf	PORTB

	movlw	high xWrite
	movwf	PCLATH
	bcf	STATUS,		C
	rlf	xPos,		F
	rlf	xPos,		W
	addwf	PCL,		F

	bsf	PORTB,		4
	bsf	PORTB,		7
	rrf	xPos,		F
	return

	bsf	PORTB,		5
	bsf	PORTB,		7
	rrf	xPos,		F
	return

	bsf	PORTB,		5
	bsf	PORTB,		6
	rrf	xPos,		F
	return

	bsf	PORTB,		4
	bsf	PORTB,		6
	rrf	xPos,		F
	return


yWrite
	movf	PORTB,		W
	andlw	0xf0
	movwf	PORTB

	movlw	high yWrite
	movwf	PCLATH
	bcf	STATUS,		C
	rlf	yPos,		F
	rlf	yPos,		W
	addwf	PCL,		F

	; sequence: 3120

	bsf	PORTB,		3
	bsf	PORTB,		0
	rrf	yPos,		F
	return

	bsf	PORTB,		1
	bsf	PORTB,		3
	rrf	yPos,		F
	return

	bsf	PORTB,		2
	bsf	PORTB,		1
	rrf	yPos,		F
	return

	bsf	PORTB,		2
	bsf	PORTB,		0
	rrf	yPos,		F
	return

zWrite
	movf	PORTD,		w
	andlw	0x0f
	movwf	PORTD

	movlw	high zWrite
	movwf	PCLATH
	bcf	STATUS,		C
	rlf	zPos,		F
	rlf	zPos,		W
	addwf	PCL,		F

	bsf	PORTD,		7
	bsf	PORTD,		5
	rrf	zPos,		F
	return

	bsf	PORTD,		4
	bsf	PORTD,		7
	rrf	zPos,		F
	return

	bsf	PORTD,		6
	bsf	PORTD,		4
	rrf	zPos,		F
	return

	bsf	PORTD,		5
	bsf	PORTD,		6
	rrf	zPos,		F
	return


xStepFwd
	incf	xPos,		W
	andlw	3
	movwf	xPos
	goto	xWrite

xStepBack
	decf	xPos,		W
	andlw	3
	movwf	xPos
	goto	xWrite

yStepFwd
	incf	yPos,		W
	andlw	3
	movwf	yPos
	goto	yWrite

yStepBack
	decf	yPos,		W
	andlw	3
	movwf	yPos
	goto	yWrite

zStepFwd
	incf	zPos,		W
	andlw	3
	movwf	zPos
	goto	zWrite

zStepBack
	decf	zPos,		W
	andlw	3
	movwf	zPos
	goto	zWrite

xRelease
	movf	PORTB,		W
	andlw	0x0f
	movwf	PORTB
	return

yRelease
	movf	PORTB,		W
	andlw	0xf0
	movwf	PORTB
	return

zRelease
	movf	PORTD,		W
	andlw	0x0f
	movwf	PORTD
	return

;; XXX remove later begin

yShowFwd
	PutChar	'y'
	return

xShowFwd
	PutChar	'x'
	return

yShowBack
	PutChar	'Y'
	return

xShowBack
	PutChar	'X'
	return

;; XXX remove later end


;;;	wait a little while so that serial comms work
waitBetweenChars
	movlw	0xFF
	movwf	ivar
iLoop
	decfsz	ivar,		F
	goto	iLoop

	retlw	0

;;;	set the ramp back to the beginning (so we'll go from slowest speed)
placeRampAtBeginning
	movlw	0x80
	movwf	period
	return

;;;	wait between steps; ramp up the speed if necessary/possible
waitBetweenSteps
	movf	period,		W
	subwf	defaultMinPer,	W
	btfsc	STATUS,		Z
	goto	waitBetweenStepsReal

	decf	period,		F

waitBetweenStepsReal
	movlw	0x40
	movwf	jvar
jLoop2
	movf 	period,		W
	movwf	ivar
iLoop2
	decfsz	ivar,		F
	goto	iLoop2

	decfsz	jvar,		F
	goto	jLoop2

	retlw	0

;;;	busy-wait until we receive a character on the UART
getChar
	bcf	PIR1,		RCIF
	bsf	RCSTA,		CREN

waitForReceive
	btfss	PIR1,		RCIF
	goto	waitForReceive

	movf	RCREG,		W
	return

;;;	output char on the UART and busy-wait for it to transmit
putChar
	movwf	TXREG
	bsf	STATUS,		RP0
waitForTransmit
	btfss	TXSTA ^ 0x80,	TRMT
	goto	waitForTransmit
	
	bcf	STATUS,		RP0
	call	waitBetweenChars
	retlw	0

;;;	output an 8-bit value as 2 hex digits
putHex
	movwf	tvar
	swapf	tvar,		W
	andlw	0xF
	call	hexDigit
	call	putChar
	movf	tvar,		W
	andlw	0xF
	call	hexDigit
	goto	putChar


org	0x300
;;;	convert a number from 0->15 to a hex digit from 0->f
hexDigit
	movwf	svar
	movlw	high hexDigit
	movwf	PCLATH
	movf	svar,		W
	addwf	PCL,		F
	retlw	'0'
	retlw	'1'
	retlw	'2'
	retlw	'3'
	retlw	'4'
	retlw	'5'
	retlw	'6'
	retlw	'7'
	retlw	'8'
	retlw	'9'
	retlw	'a'
	retlw	'b'
	retlw	'c'
	retlw	'd'
	retlw	'e'
	retlw	'f'

	end
