#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "mill.h"

int absX, absY;
int first_time;

int mirror_x = 1;
int mirror_y = 0;

double xMin;
double yMin;

#define REAL_DEVICE

#define SCALE_FACTOR 1.04

static void translate(int dx, int dy)
{
    if(mirror_x)
        dx = -dx;
    if(mirror_y)
        dy = -dy;
    mill_translate(dx, dy);
}

static void read_outline(char *file, void (*pre_line)(void),
                         void (*pt)(double, double), void (*post_line)(void))
{
    char line[65536];

    FILE *f = fopen(file, "r");
    if(!f) {
        fprintf(stderr, "cannot open '%s'\n", file);
        return;
    }

    while(fgets(line, sizeof(line), f)) {
        char *sp;

        line[strlen(line) - 1] = '\0';

        if(memcmp(line, "layer", 5)==0) continue;
        if(memcmp(line, "set", 3)==0) continue;
        if(memcmp(line, "wire", 4) != 0) {
            fprintf(stderr, "unknown command in script\n");
            fclose(f);
            return;
        }
        sp = line + 4;
        while(isspace(*sp)) sp++;

        if(pre_line) (*pre_line)();
        while(strlen(sp)) {
            double xMm, yMm;
    
            if(sscanf(sp, "(%lf %lf)", &xMm, &yMm) != 2) break;
#ifdef REAL_DEVICE
            // don't know the source of this error (leadscrew? steppers?) but
            // have to correct for it (measured from failed board)
            xMm /= SCALE_FACTOR;
            yMm /= SCALE_FACTOR;
#endif
            while(*sp != ')') sp++;
            sp++;
            while(isspace(*sp)) sp++;
    
            if(pt) (*pt)(xMm, yMm);
        }

        if(post_line) (*post_line)();
    }

    fclose(f);
}

static void real_line(void)
{
    first_time = 1;
}

static void real_post_line(void)
{
    mill_up();
}

static void real_point(double xMm, double yMm)
{
    int x, y;

    xMm -= xMin;
    yMm -= yMin;

    x = (int)(xMm * (1000 / 25.4));
    y = (int)(yMm * (1000 / 25.4));

    translate(x - absX, y - absY);

    absX = x;
    absY = y;

    if(first_time) {
        mill_down();
        first_time = 0;
    }
}

static void buildmin_point(double xMm, double yMm)
{
    if(xMm < xMin) xMin = xMm;
    if(yMm < yMin) yMin = yMm;
}

static void read_excellon(char *file)
{
    FILE *f;
    char ex_file[1024];
    char line[1024];

    int first_time;

    strcpy(ex_file, file);
    memcpy(ex_file + strlen(ex_file) - 3, "drd", 3);
    f = fopen(ex_file, "r");
    if(!f) {
        fprintf(stderr, "couldn't get excellon file '%s'\n", ex_file);
        return;
    }

    first_time = 1;

    while(fgets(line, sizeof(line), f)) {
        int x, y;
        if(*line == '%') continue;
        if(*line == 'T') {
            mill_nextsize();
            continue;
        }
        if(sscanf(line, "X%dY%d", &x, &y) != 2) {
            fprintf(stderr, "bad format: '%s'\n", line);
            fclose(f);
            return;
        }
#ifdef REAL_DEVICE
        x = (int)(x / SCALE_FACTOR);
        y = (int)(y / SCALE_FACTOR);
#endif

        if(first_time) {
            absX = x;
            absY = y;
            first_time = 0;
        }
        translate(x - absX, y - absY);
        mill_drill();
        absX = x;
        absY = y;
    }
}

void mill_from_file(char *file)
{
    xMin = 30000;
    yMin = 30000;

    absX = 0;
    absY = 0;

    read_outline(file, NULL, buildmin_point, NULL);
    read_outline(file, real_line, real_point, real_post_line);

    mill_now_drilling();

    read_excellon(file);
}
